-- By Simon de Leon, Brian Chan 3/22/05
--
-- ECE532H1S - Digital Hardware
--
--	Testbench for FFT controller
-- Sends control signals over OPB bus to configure FFT and intiate calculations,
-- and sends/receives data across FSL link.  
-- 
-- Notes: 
-- 1)  FSL and OPB model have not been instantiated in this testbench.  Latencies
-- inferred when interfacing with FSL and OPB busses below are result of h/w 
-- debug and are accurate.
-- 2)  Testbench is not self-checking b/c of scaling and overflow issues.

LIBRARY IEEE;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.STD_LOGIC_ARITH.ALL;
USE IEEE.STD_LOGIC_UNSIGNED.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE IEEE.STD_LOGIC_TEXTIO.ALL;
USE STD.TEXTIO.ALL;

ENTITY tb_fft_controller is
end tb_fft_controller;

ARCHITECTURE behavior OF tb_fft_controller is 
	-----------------------------------------------------------------------------
	-- Component Declarations 
	-----------------------------------------------------------------------------

	component fft
	port(
		OPB_Clk : IN std_logic;
		OPB_Rst : IN std_logic;
		OPB_ABus : IN std_logic_vector(0 to 31);
		OPB_BE : IN std_logic_vector(0 to 3);
		OPB_RNW : IN std_logic;
		OPB_select : IN std_logic;
		OPB_seqAddr : IN std_logic;
		OPB_DBus : IN std_logic_vector(0 to 31);
		FSL_M_Full : IN std_logic;
		FSL_S_Data : IN std_logic_vector(0 to 31);
		FSL_S_Control : IN std_logic;
		FSL_S_Exists : IN std_logic;          
		OPB_FFT_CONTROLLER_DBus : OUT std_logic_vector(0 to 31);
		OPB_FFT_CONTROLLER_errAck : OUT std_logic;
		OPB_FFT_CONTROLLER_retry : OUT std_logic;
		OPB_FFT_CONTROLLER_toutSup : OUT std_logic;
		OPB_FFT_CONTROLLER_xferAck : OUT std_logic;
		FSL_M_Clk : OUT std_logic;
		FSL_M_Write : OUT std_logic;
		FSL_M_Data : OUT std_logic_vector(0 to 31);
		FSL_M_Control : OUT std_logic;
		FSL_S_Clk : OUT std_logic;
		FSL_S_Read : OUT std_logic
		);
	end component;

	component sinusoid IS
		port (
		THETA: IN std_logic_VECTOR(9 downto 0);
		CLK: IN std_logic;
		SCLR: IN std_logic;
		COSINE: OUT std_logic_VECTOR(15 downto 0));
	END component;

	-----------------------------------------------------------------------------
	-- Signal Declarations 
	-----------------------------------------------------------------------------
	signal clk		: std_logic; -- this clock leads OPB_Clk, used for data setup
	signal OPB_Clk :  std_logic;
	signal OPB_Rst :  std_logic;
	signal OPB_ABus :  std_logic_vector(0 to 31);
	signal OPB_BE :  std_logic_vector(0 to 3);
	signal OPB_RNW :  std_logic;
	signal OPB_select :  std_logic;
	signal OPB_seqAddr :  std_logic;
	signal OPB_DBus :  std_logic_vector(0 to 31);
	signal OPB_FFT_CONTROLLER_DBus :  std_logic_vector(0 to 31);
	signal OPB_FFT_CONTROLLER_errAck :  std_logic;
	signal OPB_FFT_CONTROLLER_retry :  std_logic;
	signal OPB_FFT_CONTROLLER_toutSup :  std_logic;
	signal OPB_FFT_CONTROLLER_xferAck :  std_logic;
	signal FSL_M_Clk :  std_logic;
	signal FSL_M_Write :  std_logic;
	signal FSL_M_Data :  std_logic_vector(0 to 31);
	signal FSL_M_Control :  std_logic;
	signal FSL_M_Full :  std_logic;
	signal FSL_S_Clk :  std_logic;
	signal FSL_S_Read :  std_logic;
	signal FSL_S_Data :  std_logic_vector(0 to 31);
	signal FSL_S_Control :  std_logic;
	signal FSL_S_Exists :  std_logic;
	signal result : real := 0.0;
	signal result_int : integer;
	signal scale : real;

	-- audio generator signals
	signal theta			: std_logic_vector(9 downto 0);
	signal cosine			: std_logic_vector(15 downto 0);
	signal cosine_i			: std_logic_vector(15 downto 0);
  	signal ramp				: std_logic_vector(0 to 15);
	signal ramp_i				: std_logic_vector(0 to 15);
	signal clear			: std_logic;

	-- audio representation signals
	signal left_in				: std_logic_vector(15 downto 0);
	signal right_in				: std_logic_vector(15 downto 0);
	signal left_out				: std_logic_vector(15 downto 0);
	signal right_out				: std_logic_vector(15 downto 0);

	-- operands is used to loop-back the coefficients from the previous
	-- forward FFT into the subsequent IFFT to see if we can recover the audio
	type word is array (0 to 31) of std_logic; -- fft operand
	type wordarray is array(2048 downto 0) of std_logic_vector(0 to 31) ;
	signal operands		: wordarray;
	signal k					: integer := 0;

	-----------------------------------------------------------------------------
	-- Constant Declarations
	-----------------------------------------------------------------------------
	constant half_period		: time := 50 ns;
	constant setup				: time := 10 ns;	

	-----------------------------------------------------------------------------
	-- Functions
	-----------------------------------------------------------------------------
   function vector_to_int(sig : std_logic_vector(15 downto 0)) return integer is
   	variable num : integer := 0;  -- descending sig as integer
	begin
   	for i in sig'range loop
      	if sig(i)='1' then
        		num := num*2+1;
      	else  -- use anything other than '1' as '0'
        		num := num*2;
      	end if;
    	end loop;
		
		if sig(15)='1' then
			num := num - (2*32768);
		end if;
		return num;
  	end function vector_to_int;

begin

	-----------------------------------------------------------------------------
	-- Instantiations
	-----------------------------------------------------------------------------
	uut: fft port map(
		OPB_Clk => OPB_Clk,
		OPB_Rst => OPB_Rst,
		OPB_ABus => OPB_ABus,
		OPB_BE => OPB_BE,
		OPB_RNW => OPB_RNW,
		OPB_select => OPB_select,
		OPB_seqAddr => OPB_seqAddr,
		OPB_DBus => OPB_DBus,
		OPB_FFT_CONTROLLER_DBus => OPB_FFT_CONTROLLER_DBus,
		OPB_FFT_CONTROLLER_errAck => OPB_FFT_CONTROLLER_errAck,
		OPB_FFT_CONTROLLER_retry => OPB_FFT_CONTROLLER_retry,
		OPB_FFT_CONTROLLER_toutSup => OPB_FFT_CONTROLLER_toutSup,
		OPB_FFT_CONTROLLER_xferAck => OPB_FFT_CONTROLLER_xferAck,
		FSL_M_Clk => FSL_M_Clk,
		FSL_M_Write => FSL_M_Write,
		FSL_M_Data => FSL_M_Data,
		FSL_M_Control => FSL_M_Control,
		FSL_M_Full => FSL_M_Full,
		FSL_S_Clk => FSL_S_Clk,
		FSL_S_Read => FSL_S_Read,
		FSL_S_Data => FSL_S_Data,
		FSL_S_Control => FSL_S_Control,
		FSL_S_Exists => FSL_S_Exists
	);

	sinusoid_0: sinusoid
		port map (
		THETA => theta,
		CLK => clk,
		SCLR => OPB_Rst,
		COSINE => cosine_i
	);

	audio_representations:
	for i in 0 to 15 generate
			left_in(i) <= FSL_S_Data(i);
			right_in(i) <= FSL_S_Data(i+16);
			left_out(i) <= FSL_M_Data(i);
			right_out(i) <= FSL_M_Data(i+16);
	end generate;

	-----------------------------------------------------------------------------
	-- User Testbench
	-----------------------------------------------------------------------------

	clk_generate : process -- clock process for clk,
	begin
		clock_loop : loop
			clk <= transport '1';
			wait for half_period;
			clk <= '0';
			wait for half_period;
		end loop;			
	end process;

	OPB_Clk <= transport clk;	

	-- the incoming audio from AC97 is big-endian, but the cosine generator is little-endian.
	-- Simulate this by creating big-endian cosine and send this to the slave bus
	signed_to_unsigned : process(clk,cosine_i)
	begin
		if clk'event and clk='1' then
			for i in 0 to 15 loop
				cosine(i) <= cosine_i(15-i);
			end loop;
		end if;
	end process;	

	-- controls frequency of the sine wave
	theta_generate : process(clk)
		variable count : integer := 0;
	begin
		count := count + 20;	-- modify the addition constant to increase the frequency
		if count >= 1024 then -- theta is 2^10 max
			  count := 0;
		end if;
		if clk'event and clk='1' then
			theta <= conv_std_logic_vector(count,10);
		end if;
	end process;

	-- generates triangle wave 
	tri_generate : process (clk) -- triangle generator
		variable term 		: integer := 32767; -- centre of unsigned corresponds to 0 at xn_re
		variable flag 		: integer := 0;
	begin
		if clk'event and clk='1' then
			if flag=0 then
				term := term + 1000;
				if term >= (12000) then
					flag := 1;
				else
					flag := 0;
				end if;
			elsif flag=1 then
				term := term - 1000;
				if term <= (-12000) then
					flag := 0;
				else
					flag := 1;
				end if;
			end if;	
		end if;
		ramp_i <= conv_std_logic_vector(term,16);
		-- audio from AC97 is big-endian, but conv_std_logic_vector produces little-endian
		-- flip endianness here to simulate this, fft should handle endian conversions internally
		for i in 0 to 15 loop
			ramp(i) <= ramp_i(15-i);
		end loop;
	end process;	
																				 
	main_testbench : process			  
		procedure initialize is
		begin
			FSL_S_Data <= X"1111_1111";
			OPB_Rst <= '1';
			OPB_select<='0';
			OPB_RNW <= '0';
		  	wait for 1 us;
			OPB_Rst <= '0';
			wait for 1 us;			
		end;

		procedure fft(
			nfft 			: std_logic_vector(4 downto 0);
			in_type	 	: integer; -- 1=DC, 2=cosine, 3=triangle
			fwd_inv 		: std_logic;
			scale_sch	: std_logic_vector(11 downto 0)
		) is
			variable i 	: integer	:= 0;
			variable delay : integer :=0;
		begin
			i := 0;
			
			-- select FFT on the OPB bus, tell FFT data exists in the FSL
			wait until clk'event and clk='1';
				OPB_select <= '1';			--setup for chip select
				OPB_ABus <= X"FFFF_8000";	--sets the chip select to high
				FSL_S_Exists <= '1';			--start filling up the FSL FIFO, FIFO should do this automatically
			
			-- write in the configuration information and signal start on OPB bus
			-- the opb bus is little endian, internal fft is big endian
			wait until clk'event and clk='1';
				OPB_DBus(31) <= fwd_inv;		-- signals forward or inverse  
				for i in 0 to 4 loop
					OPB_DBus(i+16) <= nfft(i); -- s/w perspective, nfft values are reversed from used here				
				end loop;
				for i in 0 to 11 loop
					OPB_DBus(i) <= scale_sch(i);
				end loop;	
			
			k <= 0; -- initialize index for feedback of coefficients
			-- FSL presents data next edge after observing read high
			-- Loop back coefficients if doing an inverse to compare recovered audio
			wait until FSL_S_Read='1'; --starting to empty the FIFO and apply data on every clock
			wait until clk'event and clk='1'; -- takes FIFO two cycles to output data in response to read
			wait until clk'event and clk='1';
			while FSL_S_Read='1' loop
				if fwd_inv='1' then
					if in_type=1 then
						FSL_S_Data <= X"5555_5555";
					elsif in_type=2 then				
						for i in 0 to 15 loop
							FSL_S_Data(i) <= cosine(i); -- cosine is big endian, was flipped from a little endian source
						end loop;
					elsif in_type=3 then
						for i in 0 to 15 loop
							FSL_S_Data(i) <= ramp(i); -- ramp is big endian, was flipped from a little endian source
						end loop;
					end if;
					FSL_S_Data(0 to 15) <= (others =>'0'); -- no imaginary audio		
				else -- ifft the results from last forward fft
					FSL_S_Data <= operands(k);
					k <= k+1;
				end if;
				wait until clk'event and clk='1';				
			end loop;
			k<=0; -- resent index for feedback of coefficients

			--set the FIFO as empty when the fft has had enough!
			FSL_S_Exists <= '0'; -- this triggers the calculate state
			OPB_select <= '0';	--to allow for chip select to go high again next pulse

			-- unload and write data synchronously to FSL 
			wait until FSL_M_Write ='1';
			while FSL_M_Write='1' loop
				wait until clk'event and clk='1';
					operands(k) <= FSL_M_Data;
					k <= k+1;
			end loop;	

			wait for 100 ns;
		end;	
	begin
		initialize;
	  	-----------------------------------------------------------------------------
		-- N-Point FFT Tests
		-- Forward FFT is always followed by an IFFT that uses the coefficients from 
		-- previous operation to see if we can recover audio.  Scaling schedule  and
		-- amplitude of input plays large role in accuracy of recovered audio (overflow, 
		-- etc.). 
		--
		-- Notes: Arguments to fft procedure below are little-endian, the datasheet table 4
		-- is little-endian, and the microblaze commands sent are big-endian. 
		-----------------------------------------------------------------------------

		--64-point fft tests
			-- DC input
			-- unsigned cosine input
			-- unsigned triangle input
		for i in 1 to 3 loop
			fft("00110", i, '1', "101010101010");	
			fft("00110", i, '0', "101010101010");
		end loop;

		--128-point fft tests
			-- DC input
			-- unsigned cosine input
			-- unsigned triangle input
--		for i in 1 to 3 loop
--			fft("00111", i, '1', "101010101010");	
--			fft("00111", i, '0', "101010101010");
--		end loop;

		--256-point fft tests
			-- DC input
			-- unsigned cosine input
			-- unsigned triangle input
--		for i in 1 to 3 loop
--			fft("01000", i, '1', "101010101010");	
--			fft("01000", i, '0', "101010101010");
--		end loop;

		--512-point fft tests
			-- DC input
			-- unsigned cosine input
			-- unsigned triangle input
--		for i in 1 to 3 loop
--			fft("01001", i, '1', "101010101010");	
--			fft("01001", i, '0', "101010101010");
--		end loop;

		--1024-point fft tests
			-- DC input
			-- unsigned cosine input
			-- unsigned triangle input
--		for i in 1 to 3 loop
--			fft("01010", i, '1', "101010101010");	
--			fft("01010", i, '0', "101010101010");
--		end loop;

		--2048-point fft tests
			-- DC input
			-- unsigned cosine input
			-- unsigned triangle input
--		for i in 1 to 3 loop
--			fft("01011", i, '1', "101010101010");	
--			fft("01011", i, '0', "101010101010");
--		end loop;
		
		wait for 1 us; 

		assert (FALSE) report
			"Testing complete - Verify values with MATLAB"	
		severity FAILURE;	

	end process;
	-- *** end Test Bench - User Defined Section ***		
end;
