-- Simon de Leon, 3/22/05
-- Handles the FFT core

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

--  Uncomment the following lines to use the declarations that are
--  provided for instantiating Xilinx primitive components.
library unisim;
use unisim.all;

entity FFT is
	generic(
		C_OPB_AWIDTH      : integer                   	:= 32;
	   C_OPB_DWIDTH      : integer                   	:= 32;
	   C_BASEADDR        : std_logic_vector(0 to 31) 	:= X"FFFF_8000";
	   C_HIGHADDR        : std_logic_vector          	:= X"FFFF_80FF";
    	C_FSL_DWIDTH		: integer							:= 32);	
	port (
		-- Global signals
	   OPB_Clk : in std_logic;
	   OPB_Rst : in std_logic;

	   -- OPB signals
	   OPB_ABus    : in std_logic_vector(0 to C_OPB_AWIDTH-1);
	   OPB_BE      : in std_logic_vector(0 to 3);
	   OPB_RNW     : in std_logic;
	   OPB_select  : in std_logic;
	   OPB_seqAddr : in std_logic;
	   OPB_DBus    : in std_logic_vector(0 to C_OPB_DWIDTH-1);

	   OPB_FFT_CONTROLLER_DBus    : out std_logic_vector(0 to C_OPB_DWIDTH-1);
	   OPB_FFT_CONTROLLER_errAck  : out std_logic;
	   OPB_FFT_CONTROLLER_retry   : out std_logic;
	   OPB_FFT_CONTROLLER_toutSup : out std_logic;
	   OPB_FFT_CONTROLLER_xferAck : out std_logic;	
		 	 
		--FSL Master Signals (Output Results) 	
    	FSL_M_Clk     		: out std_logic;
     	FSL_M_Write   		: out std_logic;
     	FSL_M_Data    		: out std_logic_vector(0 to C_FSL_DWIDTH-1);
     	FSL_M_Control 		: out std_logic;
     	FSL_M_Full    		: in  std_logic;

		--FSL Slave Signals (Input Operands)
		FSL_S_Clk     	: out std_logic;
     	FSL_S_Read    	: out std_logic;
     	FSL_S_Data    	: in  std_logic_vector(0 to C_FSL_DWIDTH-1);
     	FSL_S_Control 	: in  std_logic;
     	FSL_S_Exists  	: in  std_logic);
end FFT;

architecture Behavioral of FFT is

	-----------------------------------------------------------------------------
	-- Component Declarations
	-----------------------------------------------------------------------------
	component fft_2048 is
		port (
		xn_re: IN std_logic_VECTOR(15 downto 0);
		xn_im: IN std_logic_VECTOR(15 downto 0);
		start: IN std_logic;
		unload: IN std_logic;
		nfft: IN std_logic_VECTOR(4 downto 0);
		nfft_we: IN std_logic;
		fwd_inv: IN std_logic;
		fwd_inv_we: IN std_logic;
		scale_sch: IN std_logic_VECTOR(11 downto 0);
		scale_sch_we: IN std_logic;
		sclr: IN std_logic;
		clk: IN std_logic;
		xk_re: OUT std_logic_VECTOR(15 downto 0);
		xk_im: OUT std_logic_VECTOR(15 downto 0);
		xn_index: OUT std_logic_VECTOR(10 downto 0);
		xk_index: OUT std_logic_VECTOR(10 downto 0);
		rfd: OUT std_logic;
		busy: OUT std_logic;
		dv: OUT std_logic;
		edone: OUT std_logic;
		done: OUT std_logic);
	end component fft_2048;

	component pselect is
	  	generic (
	      C_AB  : integer;
	      C_AW  : integer;
	      C_BAR : std_logic_vector);
	   port (
	      A      : in  std_logic_vector(0 to C_AW-1);
	      AValid : in  std_logic;
	      CS     : out std_logic);
	end component pselect;
	
	component loopback is
		port (
			addr: IN std_logic_VECTOR(10 downto 0);
			clk: IN std_logic;
			din: IN std_logic_VECTOR(31 downto 0);
			dout: OUT std_logic_VECTOR(31 downto 0);
			sinit: IN std_logic;
			we: IN std_logic);
	end component loopback;

  component FDR is
    port (Q : out std_logic;
          C : in  std_logic;
          D : in  std_logic;
          R : in  std_logic);
  end component FDR;
	-----------------------------------------------------------------------------
	-- Signal Declarations 
	-----------------------------------------------------------------------------
	--Global Signals
   signal clk							: std_logic;
	signal rst							: std_logic;

	-- Internal OPB signals
	signal OPB_ABus_i    				:  std_logic_vector(0 to C_OPB_AWIDTH-1);
	signal OPB_RNW_i     				:  std_logic;
	signal OPB_select_i  				:  std_logic;
	signal OPB_seqAddr_i 				:  std_logic;
	signal OPB_DBus_i    				:  std_logic_vector(C_OPB_DWIDTH-1 downto 0);
	signal OPB_FFT_CONTROLLER_DBus_i	: 	std_logic_vector(C_OPB_DWIDTH-1 downto 0);
	signal OPB_FFT_CONTROLLER_xferAck_i : 	std_logic;

	--Internal FSL Master Signals
	signal FSL_M_Write_i 		: std_logic;
	signal FSL_M_Data_i			: std_logic_vector(C_FSL_DWIDTH-1 downto 0); 
	signal FSL_M_Control_i		: std_logic;
	signal FSL_M_Full_i			: std_logic;

	--Internal FSL Slave Signals
	signal FSL_S_Read_i 			: std_logic;
	signal FSL_S_Data_i			: std_logic_vector(C_FSL_DWIDTH-1 downto 0);
	signal FSL_S_Control_i		: std_logic;
	signal FSL_S_Exists_i		: std_logic;

	--Internal FFT Signals
	signal start				: std_logic:='0';
	signal nfft					: std_logic_vector(4 downto 0);
	signal nfft_we				: std_logic;
	signal fwd_inv				: std_logic;
	signal fwd_inv_we			: std_logic;
	signal scale_sch			: std_logic_vector(11 downto 0);
	signal scale_sch_we		: std_logic;
   signal xn_index			: std_logic_vector(10 downto 0);
	signal xk_index			: std_logic_vector(10 downto 0);
	signal rfd					: std_logic;
	signal busy					: std_logic;
	signal dv					: std_logic;		  
	signal edone				: std_logic;
 	signal done					: std_logic;		
	signal fft_controller_CS	: std_logic;
	signal delay					: std_logic;
	signal unload					: std_logic;
	signal blk_exp			: std_logic_vector(4 downto 0);

	--OPB FSM Signals
	type 	opb_state_type is (  wait_for_fft_idle, wait_for_fft_write, start_fsl_fsm, latch_fwd_inv,latch_nfft );
	signal opb_state					: opb_state_type := latch_nfft; 		
 	signal opb_start					: std_logic;
	signal start_request				: std_logic;	
	signal status_reg					: std_logic_vector(C_OPB_DWIDTH-1 downto 0);  

 	--FSL FSM Signals
	type fsl_state_type is (idle, wait_for_fft, load, calculate, wait_for_fft_write, write_out);
	signal fsl_state 				: fsl_state_type := idle;

	signal fft_output					: std_logic_vector(C_FSL_DWIDTH-1 downto 0);
	signal fft_output_x					: std_logic_vector(C_FSL_DWIDTH-1 downto 0);
	signal fft_input					: std_logic_vector(C_FSL_DWIDTH-1 downto 0);
	signal fft_input_3				: std_logic_vector(C_FSL_DWIDTH-1 downto 0);
	signal fft_input_2				: std_logic_vector(C_FSL_DWIDTH-1 downto 0);
	signal fft_input_1				: std_logic_vector(C_FSL_DWIDTH-1 downto 0);

	-- Memory Signals
	signal addr							: std_logic_vector(10 downto 0);
	signal addr_3							: std_logic_vector(10 downto 0);
	signal addr_2							: std_logic_vector(10 downto 0);
	signal addr_1							: std_logic_vector(10 downto 0);
	signal din							: std_logic_vector(31 downto 0);
	signal dout							: std_logic_vector(31 downto 0);
	signal we							: std_logic;
	signal we_3							: std_logic;
	signal we_2							: std_logic;
	signal we_1							: std_logic;
	signal dv_1							: std_logic;
	signal dv_2							: std_logic;
	signal dv_3							: std_logic;

	-- OPB Register Handling
	signal opb_fft_controller_CS_3,opb_fft_controller_CS_2,opb_fft_controller_CS_1,opb_fft_controller_CS		: std_logic; -- pipeline to pulse chip select
	signal xfer_ack					: std_logic;
	signal opb_RNW_1					: std_logic;
	-----------------------------------------------------------------------------
	-- Functions
	-----------------------------------------------------------------------------
	
	-- used to derive scaling factor for unsigned audio by comparing 2's complement
	-- input and output from FFT
   function vector_to_int(sig : std_logic_vector(15 downto 0)) return integer is
   	variable num : integer := 0;  -- descending sig as integer
	begin
   	for i in sig'range loop
      	if sig(i)='1' then
        		num := num*2+1;
      	else  -- use anything other than '1' as '0'
        		num := num*2;
      	end if;
    	end loop;
		
		if sig(15)='1' then
			num := num - (2*32768);
		end if;
		return num;
  	end function vector_to_int;

	-----------------------------------------------------------------------------
	-- Constant Declarations
	-----------------------------------------------------------------------------

begin
	-----------------------------------------------------------------------------
	-- Endian Fix - uBlaze is little-endian, internal FFT core is big-endian
	-----------------------------------------------------------------------------
	-- bits are not reversed here, but need to hardcode connections to avoid 
	-- endian reversal when interfacing to the fft
	endian_fix : 
	for i in 0 to 31 generate
		FSL_M_Data(i) <= FSL_M_Data_i(i); -- fft_output is fed by loopback or FFT
		FSL_S_Data_i(i) <= FSL_S_Data(i);		
		OPB_DBus_i(i) <= OPB_DBus(i);
	   OPB_FFT_CONTROLLER_DBus(i) <= OPB_FFT_CONTROLLER_DBus_i(i);

		--loopback connections for debugging
		--din(i) <= FSL_S_Data(i);
		--FSL_M_Data(i) <= dout(i);
		--FSL_M_Data(i) <= fft_output(i);
	end generate;
	
	-----------------------------------------------------------------------------
	-- Combinatorial Handling
	-----------------------------------------------------------------------------
	--Global Connections
	clk <= OPB_Clk;
	FSL_M_Clk <= clk;	
	FSL_S_Clk <= clk;
	rst <= OPB_Rst;

	--Connect Internal OPB Signals
	OPB_ABus_i <= OPB_ABus;
	OPB_RNW_i <= OPB_RNW;
	OPB_select_i <= OPB_select;
	OPB_seqAddr_i <= OPB_seqAddr;
	OPB_FFT_CONTROLLER_xferAck <= OPB_FFT_CONTROLLER_xferAck_i; 

	--Connect Internal FSL Master Signals
	FSL_M_Write <= FSL_M_Write_i;
	FSL_M_Control <= FSL_M_Control_i;
	FSL_M_Full_i <= FSL_M_Full;

	--Connect Internal FSL Slave Signals
	FSL_S_Read <= FSL_S_Read_i; 
	FSL_S_Control_i <= FSL_S_Control;	
	FSL_S_Exists_i <= FSL_S_Exists;	

	-----------------------------------------------------------------------------
	-- Handling the FSL bus Interface
	-----------------------------------------------------------------------------
	--FSL Master/Slave Combinatorial
	FSL_M_Control_i <= '0'; -- always write data back onto FSL, never control data
	FSL_M_Write_i <= dv_1;	  -- use this when taking out LOOPBACK
	--FSL_M_Write_i <= dv_3;	-- use this when using LOOPBACK (want to begin write after FFT 4 cycle delay)
	FSL_S_Read_i <= rfd; -- begin reading coincident with rfd, do not delay by a register


	--FSL FSM  (idle, wait_for_fft_read, load, calculate, wait_for_fft_write, write_out)
	fsl_fsm : process(clk, rst)
		variable count		: integer :=0;
	begin
		if rst = '1' then
			fsl_state <= idle;
		elsif clk'event and clk = '1' then
			case fsl_state is
				when idle =>
					if opb_start = '1' and FSL_S_Exists_i = '1' and busy = '0' and rfd='0' then -- rfd check to prevent interfering with previous burst
						start <= '1'; -- pulse start for only 1 clock cycle to avoid multiple burst reads
						fsl_state <= wait_for_fft;	 
					else
						fsl_state <=	idle;
					end if;
			  	when wait_for_fft =>
			  		start <= '0'; -- pulse start for only 1 clock cycle to avoid multiple burst reads
					if rfd = '1' then
						fsl_state <= load;
					else
						fsl_state <= wait_for_fft;
					end if;
				when load =>
					if FSL_S_Exists_i = '0' then -- make sure the FIFO is empty, set this in testbench!
						fsl_state <= calculate;
					else
						fsl_state <= load;
					end if;
				when calculate =>
					if done = '1' then
						fsl_state <= wait_for_fft_write;
					else
						fsl_state <= calculate;
					end if;
				when wait_for_fft_write =>
					if dv = '1' then -- FSL_M_Write_i occurs simultaneously with dv, combinatorial assignment 
						fsl_state <= write_out;
					else 
						fsl_state <= wait_for_fft_write;
					end if;
				when write_out =>
					if dv = '0' then -- wait until fft empty before resetting
						fsl_state <= idle;
					else
						fsl_state <= write_out;
					end if;	
			end case;
		end if;							
	end process;

	-- converts unsigned audio to 2's complement on forward FFT and 
	-- pipes input to arrive 4 cycles later than corresponding index (see datasheet)
	unsigned_to_signed : process(clk,rst) -- delay input by 4 cycles required by fft
	begin
		if rst='1' then
			fft_input <= (others=>'0');
		elsif clk'event and clk='1' then
			if fwd_inv='1' then -- 16-bit unsigned audio input, flip endian
				fft_input(15 downto 0) <= (others=>'0'); -- imaginary part
				for i in 0 to 15 loop
					fft_input(i+16) <= fft_input_2(31-i);
				end loop;
			else -- 32-bit coefficients
				--fft_input <= fft_input_2;
				for i in 0 to 15 loop 
					fft_input(i) <= fft_input_2(15-i); -- flip endianness of imaginary input
					fft_input(i+16) <= fft_input_2(31-i); -- flip endianness of real input
				end loop;
			end if;
			  fft_input_2 <= FSL_S_Data_i;
		end if;		
	end process;

	signed_to_unsigned : process(rst, clk)
	begin
		if rst='1' then
			FSL_M_Data_i <= (others=>'0');	
		elsif clk'event and clk='1' then
			if fwd_inv='1' then
		  		--FSL_M_Data_i <= fft_output;	
				for i in 0 to 15 loop
					FSL_M_Data_i(i) <= fft_output(15-i); -- flip endianness of imaginary output
					FSL_M_Data_i(i+16) <= fft_output(31-i); -- flip endianness of real output
				end loop;
			elsif fwd_inv='0' then	-- recovering audio, flip endian
				FSL_M_Data_i(15 downto 0) <= (others=>'0');
				for i in 0 to 15 loop
					FSL_M_Data_i(i+16) <= fft_output(31-i);
				end loop;				
			end if;
		end if;
	end process;

	-----------------------------------------------------------------------------
	-- Handling the OPB bus interface
	-----------------------------------------------------------------------------

	--OPB Combinatorial
	OPB_FFT_CONTROLLER_errAck  <= '0';
	OPB_FFT_CONTROLLER_retry   <= '0';
	OPB_FFT_CONTROLLER_toutSup <= '0';
   OPB_FFT_CONTROLLER_xferAck_i <= xfer_Ack;
				 
	opb_fsm : process(clk, rst)
		variable count		: integer :=0;
	begin
		if rst = '1' then
			nfft_we <= '0';
			fwd_inv_we <= '0';
			scale_sch_we <= '0';
			opb_start <= '0';
		elsif clk'event and clk = '1' then
			case opb_state is
				when latch_nfft =>
					if xfer_Ack='1' then
						nfft_we <= '1';
						fwd_inv_we <= '0'; -- latch fwd_inv one cycle later like FFT datasheet
						scale_sch_we <= '0';
						opb_state <= latch_fwd_inv;
					else
						opb_state <= latch_nfft;
					end if;
				when latch_fwd_inv =>
					nfft_we <= '0';
					fwd_inv_we <= '1'; -- latch fwd_inv one cycle later like FFT datasheet
					scale_sch_we <= '1';
					opb_state <= start_fsl_fsm;					
				when start_fsl_fsm =>
					nfft_we <= '0';
					fwd_inv_we <= '0';	
					scale_sch_we <= '0';
					opb_start <= '1';	
					opb_state <= wait_for_fft_write;
				when wait_for_fft_write =>
					opb_start <= '0';	
					if fsl_state=write_out then
						opb_state <= wait_for_fft_idle;
					else
						opb_state <= wait_for_fft_write;
					end if;
				when wait_for_fft_idle =>
					if fsl_state=idle then
						opb_state <= latch_nfft;
					else
						opb_state <= wait_for_fft_idle;
					end if;					
			end case;
		end if;							
	end process;


  FFT_Write_Reg_Data : process (OPB_Clk, OPB_Rst) is
  begin  -- process AC97_Write_Reg_Data
    if OPB_Rst = '1' then               -- asynchronous reset (active high)
			nfft(4 downto 0) <= (others=>'0');
			fwd_inv <= '0'; 
			scale_sch(11 downto 0) <= (others=>'0');
			status_reg <= (others=>'0');
    elsif OPB_Clk'event and OPB_Clk = '1' then  -- rising clock edge
      if (opb_fft_controller_CS_2 = '1') and (OPB_RNW_1 = '0') and opb_state=latch_nfft then
			-- latch values for operation
			nfft(4 downto 0) <= OPB_DBus_i(20 downto 16);
			fwd_inv <= OPB_DBus_i(31); 
			scale_sch <= OPB_DBus_i(11 downto 0);
			status_reg <= OPB_DBus_i(C_OPB_DWIDTH-1 downto 0);
      end if;
    end if;
  end process;

  opb_fft_controller_CS_2_DFF : process (OPB_Clk, OPB_Rst) is
  begin  -- process opb_fft_controller_CS_2_DFF
    if OPB_Rst = '1' then               -- asynchronous reset (active high)
      opb_fft_controller_CS_2 <= '0';
      opb_fft_controller_CS_3 <= '0';
      opb_RNW_1                <= '0';
    elsif OPB_Clk'event and OPB_Clk = '1' then  -- rising clock edge
      opb_fft_controller_CS_2 <= opb_fft_controller_CS_1
                                  and not opb_fft_controller_CS_2
                                  and not opb_fft_controller_CS_3;
      opb_fft_controller_CS_3 <= opb_fft_controller_CS_2;
      opb_RNW_1                <= OPB_RNW;
    end if;
  end process;

  XFER_Control : process (OPB_Clk, OPB_Rst) is
  begin  -- process XFER_Control
    if OPB_Rst = '1' then               -- asynchronous reset (active high)
      xfer_Ack <= '0';
    elsif OPB_Clk'event and OPB_Clk = '1' then  -- rising clock edge
      xfer_Ack <= opb_fft_controller_CS_2;
    end if;
  end process;


	-----------------------------------------------------------------------------
	-- Loopback memory
	-----------------------------------------------------------------------------

	read_or_write : process(clk, rst)
	begin
		if rst='1' then
			we_3 <= '0';
			we_2 <= '0';
			we_1 <= '0';
			we <= '0';
			addr_3 <= (others=>'0');
			addr_2 <= (others=>'0');
			addr_1 <= (others=>'0');
			addr <= (others=>'0');
		elsif clk'event and clk='1' then	 
			dv_1 <= dv;
			dv_2 <= dv_1;
			dv_3 <= dv_2;
			we_3 <= rfd;
			we_2 <= we_3;
			we_1 <= we_2;
			we <= we_2;
			if (rfd='1' and dv='0') or we='1' then
				--delay write address to loopback memory
				addr_3 <= xn_index;
				addr_2 <= addr_3;
				addr_1 <= addr_2;
				addr <= addr_2;
			else
				-- do not delay read address to loopback memory 
				addr <= xk_index;
			end if;
		end if;			
	end process;

	-----------------------------------------------------------------------------
	-- Instantiations
	-----------------------------------------------------------------------------
	fft_2048_I : fft_2048
		port map (
			xn_re => fft_input(C_FSL_DWIDTH-1 downto C_FSL_DWIDTH/2),--fft_input(C_FSL_DWIDTH/2-1 downto 0),--fft_input(C_FSL_DWIDTH/2-1 downto 0),
			xn_im => fft_input(C_FSL_DWIDTH/2-1 downto 0),--fft_input(C_FSL_DWIDTH-1 downto C_FSL_DWIDTH/2),--fft_input(C_FSL_DWIDTH-1 downto C_FSL_DWIDTH/2),
			start => start,
			unload => done, -- begin reading results immediately
			nfft => nfft,
			nfft_we => nfft_we,
			fwd_inv => fwd_inv,
			fwd_inv_we => fwd_inv_we,
			scale_sch =>  scale_sch,
			scale_sch_we => scale_sch_we,
			sclr => rst,
			clk => clk,
			xk_re => fft_output(C_FSL_DWIDTH-1 downto C_FSL_DWIDTH/2),--fft_output(C_FSL_DWIDTH/2-1 downto 0),--fft_output(C_FSL_DWIDTH/2-1 downto 0),
			xk_im => fft_output(C_FSL_DWIDTH/2-1 downto 0),--fft_output(C_FSL_DWIDTH-1 downto C_FSL_DWIDTH/2),--fft_output(C_FSL_DWIDTH-1 downto C_FSL_DWIDTH/2),
			xn_index => xn_index,
			xk_index => xk_index,
			rfd => rfd,
			busy => busy,
			dv => dv,
			edone => open,
			done => done);

	-- Do the OPB address decoding
	pselect_I : pselect
	   generic map (
			C_AB  => C_OPB_AWIDTH,          
	      C_AW  => C_OPB_AWIDTH,          
	      C_BAR => C_BASEADDR)            
	   port map (									
	      A      => OPB_ABus_i,            
	      AValid => OPB_select_i,          
	      CS     => opb_fft_controller_CS);  

  opb_fft_controller_CS_1_DFF : FDR
    port map (
      Q => opb_fft_controller_CS_1,   
      C => OPB_Clk,                    
      D => OPB_fft_controller_CS,    
      R => xfer_Ack);                  

--	loopback_I : loopback
--		port map (
--			addr 	=>	addr,
--			clk => clk,
--			din => open,--fft_input,
--			dout => open,--fft_output,--dout,
--			sinit => rst,
--			we => we);

end architecture Behavioral;
 